#include <cmath> 
#include <cassert> 
#include <cstdio>

#define _WIN32_WINNT 0x0500
#define WINVER 0x0500

#define D3D_OVERLOADS
#include <windows.h> 
#include <atlbase.h>
#include <atlwin.h>
#include <atltypes.h>
#include <ddraw.h> 
#include <d3d.h> 
#include <comdef.h>

#pragma comment( lib, "ddraw.lib" )
#pragma comment( lib, "dxguid.lib" )

_COM_SMARTPTR_TYPEDEF(IDirectDraw, IID_IDirectDraw);
_COM_SMARTPTR_TYPEDEF(IDirectDrawSurface, IID_IDirectDrawSurface);
_COM_SMARTPTR_TYPEDEF(IDirectDrawPalette, IID_IDirectDrawPalette);
_COM_SMARTPTR_TYPEDEF(IDirectDrawClipper, IID_IDirectDrawClipper);
_COM_SMARTPTR_TYPEDEF(IDirect3D, IID_IDirect3D);
_COM_SMARTPTR_TYPEDEF(IDirect3DDevice, IID_IDirect3DDevice);
_COM_SMARTPTR_TYPEDEF(IDirect3DMaterial, IID_IDirect3DMaterial);
_COM_SMARTPTR_TYPEDEF(IDirect3DViewport, IID_IDirect3DViewport);
_COM_SMARTPTR_TYPEDEF(IDirect3DLight, IID_IDirect3DLight);
_COM_SMARTPTR_TYPEDEF(IDirect3DExecuteBuffer, IID_IDirect3DExecuteBuffer);

template< typename T>
inline void ZeroMemoryAndSetSize( T& target )
{
	ZeroMemory( &target, sizeof(T) );
	target.dwSize = sizeof(T);
}

static inline DWORD BitDepthToFlags(DWORD bitDepth) 
{ 
    switch (bitDepth) 
    { 
    case  1: return DDBD_1; 
    case  2: return DDBD_2; 
    case  4: return DDBD_4; 
    case  8: return DDBD_8; 
    case 16: return DDBD_16; 
    case 24: return DDBD_24; 
    case 32: return DDBD_32; 
    default: return 0; 
    } 
} 

static inline DWORD FlagsToBitDepth(DWORD dwFlags) 
{ 
    if (dwFlags & DDBD_1) 
        return 1; 
    else if (dwFlags & DDBD_2) 
        return 2; 
    else if (dwFlags & DDBD_4) 
        return 4; 
    else if (dwFlags & DDBD_8) 
        return 8; 
    else if (dwFlags & DDBD_16) 
        return 16; 
    else if (dwFlags & DDBD_24) 
        return 24; 
    else if (dwFlags & DDBD_32) 
        return 32; 
    else 
        return 0; 
} 

static void SetPerspectiveProjection(D3DMATRIX& matrix, D3DVALUE dHalfHeight, D3DVALUE dFrontClipping, D3DVALUE dBackClipping) 
{
    const D3DVALUE dTmp1 = dHalfHeight / dFrontClipping; 
    const D3DVALUE dTmp2 = dBackClipping / (dBackClipping - dFrontClipping); 

	ZeroMemory( &matrix, sizeof(matrix) );

    matrix._11 =  D3DVAL(2.0); 
    matrix._22 =  D3DVAL(2.0); 
    matrix._33 =  D3DVAL(dTmp1 * dTmp2); 
    matrix._34 =  D3DVAL(dTmp1); 
    matrix._43 =  D3DVAL(-dHalfHeight * dTmp2); 
} 

static void 
SetRotationAboutY(D3DMATRIX& matrix, D3DVALUE dAngleOfRotation) 
{ 
    const D3DVALUE dvCos = D3DVAL(cos(dAngleOfRotation)); 
    const D3DVALUE dvSin = D3DVAL(sin(dAngleOfRotation)); 

	ZeroMemory( &matrix, sizeof(matrix) );

	matrix._11 =  dvCos; 
    matrix._13 = -dvSin; 
    matrix._22 =  D3DVAL(1.0); 
    matrix._31 =  dvSin; 
    matrix._33 =  dvCos; 
    matrix._44 =  D3DVAL(1.0); 
} 

struct EnumInfo
{
    int  DeviceBitDepth;
    bool DeviceFound;
    bool UseHardware;

    GUID FoundDeviceGUID;
    char FoundDeviceDescStr[ 256 ]; // 256 ȉ
    char FoundDeviceName[ 256 ];    // 256 ȉ
    D3DDEVICEDESC FoundHardwareDeviceDesc;
    D3DDEVICEDESC FoundSoftwareDeviceDesc;

    EnumInfo() : DeviceBitDepth(0), DeviceFound(false), UseHardware(false), FoundDeviceGUID(GUID_NULL)
    {
        FoundDeviceDescStr[0] = '\0';
        FoundDeviceName[0] = '\0';

		ZeroMemoryAndSetSize( FoundHardwareDeviceDesc );
		ZeroMemoryAndSetSize( FoundSoftwareDeviceDesc );
    }
};

class DeviceInfo
{
    static HRESULT WINAPI 
        EnumDeviceCallback(LPGUID guid, 
        LPSTR           deviceDescStr, 
        LPSTR           deviceName, 
        LPD3DDEVICEDESC hardwareDeviceDesc, 
        LPD3DDEVICEDESC softwareDeviceDesc, 
        LPVOID          arg) 
    {
        EnumInfo& info = *(EnumInfo*)arg;

        const bool useHardware = (hardwareDeviceDesc->dcmColorModel != 0);
        const D3DDEVICEDESC* deviceDesc = (useHardware ? hardwareDeviceDesc : softwareDeviceDesc); 

        // rbg[xvĂ邩H
        if (0 == (deviceDesc->dwDeviceRenderBitDepth & info.DeviceBitDepth) ) 
        { 
            // vĂȂΎ
            return D3DENUMRET_OK; 
        } 

        // O[[VF[fBOT|[gĂ邩H
        {
            const DWORD caps = ( deviceDesc->dcmColorModel == D3DCOLOR_MONO ? D3DPSHADECAPS_COLORGOURAUDMONO : D3DPSHADECAPS_COLORGOURAUDRGB );
            if (!(deviceDesc->dpcTriCaps.dwShadeCaps & caps)) 
            { 
                // O[[VF[fBOȂ  
                return D3DENUMRET_OK; 
            }
        }

        info.DeviceFound = true; 
        CopyMemory(&info.FoundDeviceGUID, guid, sizeof(GUID)); 
        strcpy_s(info.FoundDeviceDescStr, deviceDescStr); 
        strcpy_s(info.FoundDeviceName, deviceName); 
        info.FoundHardwareDeviceDesc = *hardwareDeviceDesc;
        info.FoundSoftwareDeviceDesc = *softwareDeviceDesc;
        info.UseHardware = useHardware;

        // n[hEFAfoCX΁C
        // ȏ̗񋓂͕sv
        return useHardware ? D3DENUMRET_CANCEL : D3DENUMRET_OK;
    } 

    HRESULT CreateDirect3D() 
    { 
        HRESULT ret = E_FAIL; 

        IDirectDrawPtr dd1;
        ret = DirectDrawCreate(NULL, &dd1, NULL); 
        if (FAILED(ret)) 
            return ret; 

        ret = dd1->SetCooperativeLevel(m_windowHandle, DDSCL_NORMAL); 
        if (FAILED(ret)) 
            return ret; 

        m_dd = dd1;
        ret = m_dd.QueryInterface( m_d3d.GetIID(), &m_d3d );
        if (FAILED(ret)) 
            return ret; 

        return DD_OK; 
    } 

    HRESULT CreatePrimary() 
    { 
        HRESULT             hRes; 

        // vC}T[tFX̍쐬
		{
            DDSURFACEDESC surfaceDesc; 
			ZeroMemoryAndSetSize( surfaceDesc );
            surfaceDesc.dwFlags = DDSD_CAPS; 
            surfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE; 

            hRes = m_dd->CreateSurface( &surfaceDesc, &m_primSurf, NULL); 
            if (FAILED(hRes)) 
                return hRes; 
        }

        // Nbp̍쐬
        {
            IDirectDrawClipperPtr clipper;
            hRes = m_dd->CreateClipper(0, &clipper, NULL); 
            if (FAILED(hRes)) 
                return hRes; 
            hRes = clipper->SetHWnd(0, m_windowHandle); 
            if (FAILED(hRes)) 
            { 
                return hRes; 
            } 
            hRes = m_primSurf->SetClipper(clipper); 
            if (FAILED(hRes)) 
            { 
                return hRes; 
            }
        }

        // pbgJ[̏ꍇ
        // XP ɂ͕sv
        {
            DDSURFACEDESC surfaceDesc; 
			ZeroMemoryAndSetSize( surfaceDesc );
            hRes = m_primSurf->GetSurfaceDesc(&surfaceDesc); 
            if (FAILED(hRes)) 
                return hRes; 

            if (surfaceDesc.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8) 
            { 
                PALETTEENTRY colorTable[256]; 
                HDC hdc = GetDC(NULL); 
                GetSystemPaletteEntries(hdc, 0, 256, colorTable); 
                ReleaseDC(NULL, hdc); 

                // ŏƍŌ 10 F̓VXeJ[Ȃ̂ŌŒ肷
                for (int i = 0; i < 10; i++) 
                    colorTable[i].peFlags = D3DPAL_READONLY; 
                for (int i = 10; i < 246; i++) 
                    colorTable[i].peFlags = D3DPAL_FREE | PC_RESERVED; 
                for (int i = 246; i < 256; i++) 
                    colorTable[i].peFlags = D3DPAL_READONLY;

                hRes = m_dd->CreatePalette(DDPCAPS_8BIT, colorTable, &m_palette, NULL); 

                if (FAILED(hRes))
                    return hRes;

                hRes = m_primSurf->SetPalette(m_palette); 
                return hRes; 
            }
        }
        return DD_OK; 
    } 
public:
    IDirectDrawPtr        m_dd;
    IDirect3DPtr          m_d3d;
    IDirectDrawPalettePtr m_palette;
    IDirectDrawSurfacePtr m_primSurf;
    IDirectDrawSurfacePtr m_deviceSurf;
    IDirectDrawSurfacePtr m_zBufferSurf;
    IDirect3DDevicePtr    m_device;
    EnumInfo              m_enumInfo;
	HWND                  m_windowHandle;

	DeviceInfo() : m_enumInfo(), m_windowHandle(NULL)
	{

	}

	~DeviceInfo()
	{
		OnUninitialize();
		m_windowHandle = NULL;
		m_enumInfo = EnumInfo();

    	assert( !m_dd );
    	assert( !m_d3d );
    	assert( !m_palette );
    	assert( !m_primSurf );
    	assert( !m_deviceSurf );
    	assert( !m_zBufferSurf );
    	assert( !m_device );
	}

    HRESULT OnUninitialize()
    {
        OnDeleteDevice();
		if( m_primSurf ) m_primSurf.Release();
		if( m_palette ) m_palette.Release();
		if( m_d3d ) m_d3d.Release();
		if( m_dd ) m_dd.Release();

        return DD_OK;
    }

	HRESULT OnDeleteDevice()
    {
        if( m_device ) m_device.Release();
        if( m_deviceSurf ) m_deviceSurf.Release();
        if( m_zBufferSurf ) m_zBufferSurf.Release();
        return DD_OK;
    }

    HRESULT Initialize(HWND hwnd)
    {
		m_windowHandle = hwnd;

        HRESULT hRes = CreateDirect3D();
        if (FAILED(hRes)) 
            return hRes; 

        hRes = CreatePrimary();
        if (FAILED(hRes)) 
            return hRes; 

        return hRes;
    }

    HRESULT CreateDevice(DWORD width, DWORD height) 
    { 
        HRESULT       hRes = E_FAIL; 
        // EBhE[hgp̂ŁC
        // vC}T[tFCX̃rbg[xI

        DDSURFACEDESC ddsd; 
		ZeroMemoryAndSetSize( ddsd );
        hRes = m_primSurf->GetSurfaceDesc(&ddsd); 
        if (FAILED(hRes)) 
            return hRes; 

        m_enumInfo.DeviceBitDepth = BitDepthToFlags(ddsd.ddpfPixelFormat.dwRGBBitCount); 

        m_enumInfo.DeviceFound = false; 

        hRes = m_d3d->EnumDevices( EnumDeviceCallback, &m_enumInfo); 
        if (FAILED(hRes)) 
            return hRes; 

        if (!m_enumInfo.DeviceFound) 
        {  
            return DDERR_NOTFOUND; 
        } 

        const D3DDEVICEDESC& deviceDesc
			= m_enumInfo.UseHardware ? m_enumInfo.FoundHardwareDeviceDesc : m_enumInfo.FoundSoftwareDeviceDesc;
        const DWORD deviceMemType
			= m_enumInfo.UseHardware ? DDSCAPS_VIDEOMEMORY : DDSCAPS_SYSTEMMEMORY;
        const DWORD zBufferMemType
			= m_enumInfo.UseHardware ? DDSCAPS_VIDEOMEMORY : DDSCAPS_SYSTEMMEMORY;

        // sNZtH[}bg̓vC}T[tFCXƓɂȂ̂
        // w肷Kv͂Ȃ
        // TCYC^CvCcapabilities w
        {
            DDSURFACEDESC   ddsd; 
            ZeroMemoryAndSetSize( ddsd ); 
            ddsd.dwFlags        = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT; 
            ddsd.dwWidth        = width; 
            ddsd.dwHeight       = height; 
            ddsd.ddsCaps.dwCaps = DDSCAPS_3DDEVICE | DDSCAPS_OFFSCREENPLAIN | deviceMemType; 
            hRes = m_dd->CreateSurface( &ddsd, &m_deviceSurf, NULL); 
            if (FAILED(hRes)) 
                return hRes; 
        }

        if (m_palette) 
        { 
            hRes = m_deviceSurf->SetPalette(m_palette); 
            if (FAILED(hRes)) 
                return hRes; 
        } 

        if (0 != deviceDesc.dwDeviceZBufferBitDepth) 
        { 
            const DWORD zBufferBitDepth = FlagsToBitDepth(deviceDesc.dwDeviceZBufferBitDepth); 

            DDSURFACEDESC   ddsd; 
            ZeroMemoryAndSetSize( ddsd ); 
            ddsd.dwSize            = sizeof(ddsd); 
            ddsd.dwFlags           = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_ZBUFFERBITDEPTH; 
            ddsd.ddsCaps.dwCaps    = DDSCAPS_ZBUFFER | zBufferMemType; 
            ddsd.dwWidth           = width; 
            ddsd.dwHeight          = height; 
            ddsd.dwZBufferBitDepth = zBufferBitDepth;
            IDirectDrawSurfacePtr zbuffer;
            hRes = m_dd->CreateSurface( &ddsd, &zbuffer, NULL); 
            if (FAILED(hRes)) 
                return hRes; 

            // foCXɃA^b`
            hRes = m_deviceSurf->AddAttachedSurface(zbuffer); 
            if (FAILED(hRes)) 
                return hRes;

            m_zBufferSurf = zbuffer;
        } 

        m_deviceSurf.QueryInterface( m_enumInfo.FoundDeviceGUID, &m_device );
        if (FAILED(hRes)) 
            return hRes;

        return DD_OK; 
    } 

    HRESULT Restore()
    {
        HRESULT hRes = E_FAIL;

        if( m_primSurf )
        {
            hRes = m_primSurf->Restore(); 
            if (FAILED(hRes)) 
                return hRes;
        }
        if( m_zBufferSurf )
        {
            hRes = m_zBufferSurf->Restore();
            if (FAILED(hRes)) 
                return hRes;
        }
        if( m_deviceSurf )
        {
            hRes = m_deviceSurf->Restore(); 
            if (FAILED(hRes)) 
                return hRes;
        }
        return DD_OK;
    }
};


class ExecutionBufferWriter
{
	void* BufferPointer;
public:
	ExecutionBufferWriter( void* ptr ) : BufferPointer(ptr)
	{
	}
	template<typename T>
	void Write(const T& obj)
	{
		T* ptr = static_cast<T*>( BufferPointer );
		*ptr = obj;
		BufferPointer = ptr + 1;
	}
	void WriteInstruction( D3DOPCODE opcode, BYTE size, WORD count)
	{
		D3DINSTRUCTION instruction = {0};
		instruction.bOpcode = static_cast<BYTE>( opcode );
		instruction.bSize = size;
		instruction.wCount = count;
		Write( instruction );
	}
	void WriteState( D3DTRANSFORMSTATETYPE type, DWORD value )
	{
		D3DSTATE state;
		state.dtstTransformStateType = type;
		state.dwArg[0] = value;
		Write( state );
	}
	void WriteState( D3DLIGHTSTATETYPE type, DWORD value )
	{
		D3DSTATE state;
		state.dlstLightStateType = type;
		state.dwArg[0] = value;
		Write( state );
	}
	void WriteState( D3DRENDERSTATETYPE type, DWORD value )
	{
		D3DSTATE state;
		state.drstRenderStateType = type;
		state.dwArg[0] = value;
		Write( state );
	}
	void WriteProcessVertices(DWORD flags, WORD start, WORD dest, DWORD count)
	{
		D3DPROCESSVERTICES vp;
		vp.dwFlags = flags;
		vp.wStart = start;
		vp.wDest = dest;
		vp.dwCount = count;
		vp.dwReserved = 0;
		Write( vp );
	}
	void WriteTriangle( WORD v1, WORD v2, WORD v3, WORD flags )
	{
		D3DTRIANGLE triangle = {0};
		triangle.v1 = v1;
		triangle.v2 = v2;
		triangle.v3 = v3;
		triangle.wFlags = flags;
		Write( triangle );
	}
};

class Scene
{
	DeviceInfo                m_deviceInfo;
    IDirect3DMaterialPtr      m_material;
    IDirect3DMaterialPtr      m_bgMaterial;
    IDirect3DViewportPtr      m_viewport;
    IDirect3DLightPtr         m_light;
    IDirect3DExecuteBufferPtr m_exebuf;

    D3DMATRIXHANDLE           hd3dWorldMatrix; 
    D3DMATRIXHANDLE           hd3dViewMatrix; 
    D3DMATRIXHANDLE           hd3dProjMatrix; 
    D3DMATERIALHANDLE         hd3dSurfaceMaterial; 
    D3DMATERIALHANDLE         hd3dBackgroundMaterial;

    D3DMATRIX                 d3dWorldMatrix;
    D3DMATRIX                 d3dViewMatrix;
    D3DMATRIX                 d3dProjMatrix;

    DWORD                     Width;
    DWORD                     Height;

	static const int NUM_VERTICES          = 3;
    static const int NUM_INSTRUCTIONS      = 6;
    static const int NUM_STATES            = 8;
    static const int NUM_PROCESSVERTICES   = 1;
    static const int NUM_TRIANGLES         = 2;

	HRESULT CreateDeviceAndScene()
    {
        HRESULT              hRes = DD_OK; 

		assert( ! m_deviceInfo.m_device );

		// foCX̍쐬
		{
			hRes = m_deviceInfo.CreateDevice(Width, Height);
			if (FAILED(hRes)) 
				return hRes; 
		}

        // Cg쐬Eݒ
		{
			D3DLIGHT             d3dLight; 
			ZeroMemoryAndSetSize( d3dLight ); 
			d3dLight.dltType = D3DLIGHT_POINT; 
			d3dLight.dcvColor.dvR    = D3DVAL(1.0); 
			d3dLight.dcvColor.dvG    = D3DVAL(1.0); 
			d3dLight.dcvColor.dvB    = D3DVAL(1.0); 
			d3dLight.dcvColor.dvA    = D3DVAL(1.0); 
			d3dLight.dvPosition      = D3DVECTOR( D3DVAL( 1.0), D3DVAL(-1.0), D3DVAL(-1.0) );
			d3dLight.dvAttenuation0  = D3DVAL(1.0); 
			d3dLight.dvAttenuation1  = D3DVAL(0.1); 
			d3dLight.dvAttenuation2  = D3DVAL(0.0); 
			hRes = m_deviceInfo.m_d3d->CreateLight(&m_light, NULL); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_light->SetLight(&d3dLight); 
			if (FAILED(hRes)) 
				return hRes; 
		}

		// wi}eA쐬Eݒ
		{
			D3DMATERIAL          d3dMaterial; 
			ZeroMemoryAndSetSize( d3dMaterial ); 
			d3dMaterial.dcvDiffuse.r  = D3DVAL(0.7); 
			d3dMaterial.dcvDiffuse.g  = D3DVAL(0.7); 
			d3dMaterial.dcvDiffuse.b  = D3DVAL(0.7); 
			d3dMaterial.dcvAmbient.r  = D3DVAL(0.0); 
			d3dMaterial.dcvAmbient.g  = D3DVAL(0.0); 
			d3dMaterial.dcvAmbient.b  = D3DVAL(0.0); 
			d3dMaterial.dcvSpecular.r = D3DVAL(0.0); 
			d3dMaterial.dcvSpecular.g = D3DVAL(0.0); 
			d3dMaterial.dcvSpecular.b = D3DVAL(0.0); 
			d3dMaterial.dvPower       = D3DVAL(0.0); 
			d3dMaterial.dwRampSize    = 1; 

			hRes = m_deviceInfo.m_d3d->CreateMaterial(&m_bgMaterial, NULL); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_bgMaterial->SetMaterial(&d3dMaterial); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_bgMaterial->GetHandle(m_deviceInfo.m_device, &hd3dBackgroundMaterial); 
			if (FAILED(hRes)) 
				return hRes; 
		}

        // r[|[g̍쐬Eݒ
		{
			hRes = m_deviceInfo.m_d3d->CreateViewport(&m_viewport, NULL); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_deviceInfo.m_device->AddViewport(m_viewport); 
			if (FAILED(hRes)) 
				return hRes;
			hRes = m_viewport->SetBackground(hd3dBackgroundMaterial); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_viewport->AddLight(m_light); 
			if (FAILED(hRes)) 
				return hRes; 
	        hRes = UpdateViewport();
			if (FAILED(hRes)) 
				return hRes; 
		}

        // sIuWFNg쐬
		{
			hRes = m_deviceInfo.m_device->CreateMatrix(&hd3dWorldMatrix); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_deviceInfo.m_device->SetMatrix(hd3dWorldMatrix, &d3dWorldMatrix); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_deviceInfo.m_device->CreateMatrix(&hd3dViewMatrix); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_deviceInfo.m_device->SetMatrix(hd3dViewMatrix, 
				&d3dViewMatrix); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_deviceInfo.m_device->CreateMatrix(&hd3dProjMatrix); 
			if (FAILED(hRes)) 
				return hRes; 
			SetPerspectiveProjection(d3dProjMatrix, D3DVAL(0.5), D3DVAL(1.0) , D3DVAL(1000.0) ); 
			hRes = m_deviceInfo.m_device->SetMatrix(hd3dProjMatrix, &d3dProjMatrix); 
			if (FAILED(hRes)) 
				return hRes; 
		}

        // ̗p}eA쐬Eݒ
		{
			D3DMATERIAL          d3dMaterial; 
			ZeroMemoryAndSetSize( d3dMaterial ); 
			d3dMaterial.dcvDiffuse.r  = D3DVAL(1.0); 
			d3dMaterial.dcvDiffuse.g  = D3DVAL(1.0); 
			d3dMaterial.dcvDiffuse.b  = D3DVAL(0.0); 
			d3dMaterial.dcvAmbient.r  = D3DVAL(0.4); 
			d3dMaterial.dcvAmbient.g  = D3DVAL(0.4); 
			d3dMaterial.dcvAmbient.b  = D3DVAL(0.0); 
			d3dMaterial.dcvSpecular.r = D3DVAL(1.0); 
			d3dMaterial.dcvSpecular.g = D3DVAL(1.0); 
			d3dMaterial.dcvSpecular.b = D3DVAL(1.0); 
			d3dMaterial.dvPower       = D3DVAL(20.0); 
			d3dMaterial.dwRampSize    = 16; 
			hRes = m_deviceInfo.m_d3d->CreateMaterial(&m_material, NULL); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_material->SetMaterial(&d3dMaterial); 
			if (FAILED(hRes)) 
				return hRes; 
			hRes = m_material->GetHandle(m_deviceInfo.m_device, &hd3dSurfaceMaterial); 
			if (FAILED(hRes)) 
				return hRes; 
		}

        // sobt@̍쐬Eݒ
        {

            const DWORD dwVertexSize      = (NUM_VERTICES * sizeof(D3DVERTEX)); 
            const DWORD dwInstructionSize =
				(NUM_INSTRUCTIONS    * sizeof(D3DINSTRUCTION))     + 
                (NUM_STATES          * sizeof(D3DSTATE))           + 
                (NUM_PROCESSVERTICES * sizeof(D3DPROCESSVERTICES)) + 
                (NUM_TRIANGLES       * sizeof(D3DTRIANGLE)); 
            const DWORD dwExecuteBufferSize = dwVertexSize + dwInstructionSize; 

            D3DEXECUTEBUFFERDESC d3dExecuteBufferDesc; 
			ZeroMemoryAndSetSize( d3dExecuteBufferDesc ); 
            d3dExecuteBufferDesc.dwFlags      = D3DDEB_BUFSIZE; 
            d3dExecuteBufferDesc.dwBufferSize = dwExecuteBufferSize; 

            hRes = m_deviceInfo.m_device->CreateExecuteBuffer(&d3dExecuteBufferDesc, &m_exebuf, NULL); 
            if (FAILED(hRes)) 
                return hRes; 

			// sobt@ɖ߂ƒ_f[^
            hRes = FillExecuteBuffer(); 
            if (FAILED(hRes)) 
                return hRes; 

			D3DEXECUTEDATA       d3dExecuteData; 
			ZeroMemory(&d3dExecuteData, sizeof(d3dExecuteData)); 
			d3dExecuteData.dwSize = sizeof(d3dExecuteData); 
			d3dExecuteData.dwVertexCount       = NUM_VERTICES; 
			d3dExecuteData.dwInstructionOffset = dwVertexSize; 
			d3dExecuteData.dwInstructionLength = dwInstructionSize; 
			hRes = m_exebuf->SetExecuteData(&d3dExecuteData); 
			if (FAILED(hRes)) 
				return hRes; 
        }

        return DD_OK;
    }
	
    HRESULT FillExecuteBuffer() 
    { 
        HRESULT hRes = DD_OK; 

        D3DEXECUTEBUFFERDESC desc; 
		ZeroMemoryAndSetSize( desc ); 
        hRes = m_exebuf->Lock(&desc); 
        if (FAILED(hRes)) 
            return hRes; 

		{
			ExecutionBufferWriter ew(desc.lpData);

			// sobt@́C_zCߌQ̏ŏ܂

			// _f[^̏
			{
				const D3DVERTEX vertex1( D3DVECTOR(D3DVAL( 0.0), D3DVAL( 1.0), D3DVAL( 0.0)),
										 D3DVECTOR(D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL(-1.0)), 
										 D3DVAL(0.0), D3DVAL(1.0) );

				const D3DVERTEX vertex2( D3DVECTOR(D3DVAL( 1.0), D3DVAL(-1.0), D3DVAL( 0.0)),
										 D3DVECTOR(D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL(-1.0)), 
										 D3DVAL(1.0), D3DVAL(1.0) );

				const D3DVERTEX vertex3( D3DVECTOR(D3DVAL(-1.0), D3DVAL(-1.0), D3DVAL( 0.0)),
										 D3DVECTOR(D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL(-1.0)), 
										 D3DVAL(1.0), D3DVAL(0.0) );

				ew.Write( vertex1 );
				ew.Write( vertex2 );
				ew.Write( vertex3 );
			}

			// 疽߃f[^

			// SetTransform
			ew.WriteInstruction( D3DOP_STATETRANSFORM, sizeof(D3DSTATE), 3U );
			ew.WriteState( D3DTRANSFORMSTATE_WORLD, hd3dWorldMatrix );
			ew.WriteState( D3DTRANSFORMSTATE_VIEW, hd3dViewMatrix );
			ew.WriteState( D3DTRANSFORMSTATE_PROJECTION, hd3dProjMatrix );

			// SetLight
			ew.WriteInstruction( D3DOP_STATELIGHT, sizeof(D3DSTATE), 2U );
			ew.WriteState( D3DLIGHTSTATE_MATERIAL, hd3dSurfaceMaterial );
			ew.WriteState( D3DLIGHTSTATE_AMBIENT, RGBA_MAKE(128, 128, 128, 128) );

			// SetRenderState
 			ew.WriteInstruction( D3DOP_STATERENDER, sizeof(D3DSTATE), 3U );
			ew.WriteState( D3DRENDERSTATE_FILLMODE, D3DFILL_SOLID );
			ew.WriteState( D3DRENDERSTATE_SHADEMODE, D3DSHADE_GOURAUD );
			ew.WriteState( D3DRENDERSTATE_DITHERENABLE, TRUE );

			// _ 
			// _ϊEv
 			ew.WriteInstruction( D3DOP_PROCESSVERTICES, sizeof(D3DPROCESSVERTICES), 1U );
			ew.WriteProcessVertices( D3DPROCESSVERTICES_TRANSFORMLIGHT, 0U, 0U, NUM_VERTICES );

			// Op``
			// DrawIndexedPrimitive  PrimitiveTypeCCfbNXwɑ
 			ew.WriteInstruction( D3DOP_TRIANGLE, sizeof(D3DTRIANGLE), 2U );
			ew.WriteTriangle( 0U, 1U, 2U, D3DTRIFLAG_EDGEENABLETRIANGLE );
			ew.WriteTriangle( 2U, 1U, 0U, D3DTRIFLAG_EDGEENABLETRIANGLE );

			// I
 			ew.WriteInstruction( D3DOP_EXIT, 0, 0U );
		}

        m_exebuf->Unlock(); 

        return DD_OK; 
    } 
	
    HRESULT OnDeleteDevice() 
    { 
        m_exebuf.Release();
        hd3dBackgroundMaterial = 0;
        m_bgMaterial.Release();
        hd3dSurfaceMaterial = 0;
        m_material.Release();

        if (0 != hd3dWorldMatrix) 
        { 
            m_deviceInfo.m_device->DeleteMatrix(hd3dWorldMatrix); 
            hd3dWorldMatrix = 0; 
        } 
        if (0 != hd3dViewMatrix) 
        { 
            m_deviceInfo.m_device->DeleteMatrix(hd3dViewMatrix); 
            hd3dViewMatrix = 0; 
        } 
        if (0 != hd3dProjMatrix) 
        { 
            m_deviceInfo.m_device->DeleteMatrix(hd3dProjMatrix); 
            hd3dProjMatrix = 0; 
        } 

        m_viewport.Release();
        m_light.Release();

		m_deviceInfo.OnDeleteDevice();

        return DD_OK; 
    }
	
	HRESULT UpdateViewport() 
    { 
        D3DVIEWPORT d3dViewport; 
		ZeroMemoryAndSetSize( d3dViewport );
        d3dViewport.dwX      = 0; 
        d3dViewport.dwY      = 0; 
        d3dViewport.dwWidth  = Width; 
        d3dViewport.dwHeight = Height; 
        d3dViewport.dvScaleX = D3DVAL((float)d3dViewport.dwWidth / 2.0); 
        d3dViewport.dvScaleY = D3DVAL((float)d3dViewport.dwHeight / 2.0); 
        d3dViewport.dvMaxX   = D3DVAL(1.0); 
        d3dViewport.dvMaxY   = D3DVAL(1.0); 
        HRESULT hRes = m_viewport->SetViewport(&d3dViewport); 
        return hRes;
    } 
public:
    Scene() :
		Width(0),
		Height(0),
        d3dWorldMatrix( 
            D3DVAL( 1.0), D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 1.0), D3DVAL( 0.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 1.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 1.0) 
            ),
        d3dViewMatrix(  
            D3DVAL( 1.0), D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 1.0), D3DVAL( 0.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 1.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 5.0), D3DVAL( 1.0) 
            ),
        d3dProjMatrix( 
            D3DVAL( 2.0), D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 2.0), D3DVAL( 0.0), D3DVAL( 0.0), 
            D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL( 1.0), D3DVAL( 1.0), 
            D3DVAL( 0.0), D3DVAL( 0.0), D3DVAL(-1.0), D3DVAL( 0.0) 
            )
    {
    }

	HRESULT Initialize(HWND hwnd)
	{
		CWindow window;
		window.Attach(hwnd);

		CRect rect;
		window.GetClientRect( &rect );

		Width = rect.Width();
		Height = rect.Height();

		HRESULT hRes = DD_OK; 
		hRes = m_deviceInfo.Initialize(hwnd);
		if (FAILED(hRes)) 
			return hRes; 
		hRes = CreateDeviceAndScene();
		if (FAILED(hRes)) 
			return hRes; 

		return DD_OK;
	}

	HRESULT Uninitialize()
	{
		OnDeleteDevice();
		m_deviceInfo.OnUninitialize();

		return DD_OK;
	}

	~Scene()
	{
		Uninitialize();

    	assert( !m_material );
    	assert( !m_bgMaterial );
    	assert( !m_viewport );
    	assert( !m_light );
    	assert( !m_exebuf );

    	assert( 0 == hd3dWorldMatrix );
    	assert( 0 == hd3dViewMatrix );
    	assert( 0 == hd3dProjMatrix );
	}

    HRESULT AnimateScene() 
    {
		if( ! m_deviceInfo.m_device )
		{
			return E_FAIL;
		}

        HRESULT hRes = DD_OK;
        SetRotationAboutY(d3dWorldMatrix, GetTickCount() * 0.001f ); 
        hRes = m_deviceInfo.m_device->SetMatrix(hd3dWorldMatrix, &d3dWorldMatrix); 
        if (FAILED(hRes)) 
            return hRes; 
        return DD_OK; 
    } 

    HRESULT RenderScene() 
    { 
		if( ! m_deviceInfo.m_device )
		{
			return E_FAIL;
		}

		HRESULT hRes = DD_OK;

		D3DRECT d3dRect; 
        d3dRect.lX1 = 0; 
        d3dRect.lX2 = Width; 
        d3dRect.lY1 = 0; 
        d3dRect.lY2 = Height; 

        hRes = m_viewport->Clear(1, &d3dRect, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER); 
        if (FAILED(hRes)) 
            return hRes; 

        hRes = m_deviceInfo.m_device->BeginScene(); 
        if (FAILED(hRes)) 
            return hRes; 

        hRes = m_deviceInfo.m_device->Execute(m_exebuf, m_viewport, D3DEXECUTE_CLIPPED); 
        if (FAILED(hRes)) 
        { 
            m_deviceInfo.m_device->EndScene(); 
            return hRes; 
        } 

        hRes = m_deviceInfo.m_device->EndScene(); 
        if (FAILED(hRes)) 
            return hRes; 

        return DD_OK; 
    } 

    HRESULT DoFrame() 
    { 
		if( ! m_deviceInfo.m_device )
		{
			return E_FAIL;
		}

		while (TRUE) 
        { 
	        HRESULT hRes = RenderScene(); 
            if (SUCCEEDED(hRes)) 
            { 
                CRect srcRect(0, 0, Width, Height);

				CWindow window;
				window.Attach(m_deviceInfo.m_windowHandle);

				CRect destRect;
				window.GetClientRect( &destRect );
				window.ClientToScreen( &destRect );

                hRes = m_deviceInfo.m_primSurf->Blt( &destRect, m_deviceInfo.m_deviceSurf, &srcRect, DDBLT_WAIT, NULL); 
                if (SUCCEEDED(hRes))  // If it worked. 
                {
                    return hRes; 
                    Sleep(0);
                }
            } 
            if (DDERR_SURFACELOST == hRes)
            {
                Sleep(100);
                hRes = m_deviceInfo.Restore();
            }
            if (FAILED(hRes))
                return hRes; 
        } 
    } 
	HRESULT OnActivateApp(bool activated)
	{
        if( !m_deviceInfo.m_device )
        {
            return DD_OK;
        }
		
		if( activated && m_deviceInfo.m_primSurf && m_deviceInfo.m_palette )
		{
			HRESULT hRes = m_deviceInfo.m_primSurf->SetPalette(m_deviceInfo.m_palette); 
			return hRes;
		}
        return DD_OK;
	}
	HRESULT OnSize( DWORD width, DWORD height )
	{
        if( !m_deviceInfo.m_device )
        {
            return DD_OK;
        }
		
		if( m_deviceInfo.m_device )
        {
            DDSURFACEDESC ddsd; 
	 		ZeroMemoryAndSetSize( ddsd );
            HRESULT hRes = m_deviceInfo.m_deviceSurf->GetSurfaceDesc(&ddsd); 
            assert (SUCCEEDED(hRes));
            if ((width > ddsd.dwWidth) || (height > ddsd.dwHeight)) 
            { 
                OnDeleteDevice(); 
            } 
        }
        this->Width = width;
        this->Height = height;
        if( !m_deviceInfo.m_device )
        {
            HRESULT hRes = CreateDeviceAndScene(); 
        }
        HRESULT hRes = UpdateViewport();
        assert (SUCCEEDED(hRes));

		return DD_OK;
	}
	HRESULT OnDisplayModeChange()
	{
		if( !m_deviceInfo.m_device )
        {
            return DD_OK;
        }

		const HWND hwnd = m_deviceInfo.m_windowHandle;
		HRESULT hRes = Uninitialize();
		if (FAILED(hRes))
			return hRes; 
		hRes = Initialize(hwnd);
		if (FAILED(hRes))
			return hRes; 

        return DD_OK;
	}
};

//-----------------------------------------------------------------------------
// WindowNX
//-----------------------------------------------------------------------------
class MainWindow : public CWindowImpl<MainWindow>
{
public:
    BEGIN_MSG_MAP( MainWindow )
		MESSAGE_HANDLER( WM_ACTIVATEAPP,   OnActivateApp )
		MESSAGE_HANDLER( WM_DISPLAYCHANGE, OnDisplayModeChange )
        MESSAGE_HANDLER( WM_DESTROY,       OnDestroy )
		MESSAGE_HANDLER( WM_KEYDOWN,       OnKeyDown )
		MESSAGE_HANDLER( WM_SIZE,          OnSize )
    END_MSG_MAP()

    Scene*      m_scene;

    MainWindow() : m_scene(NULL)
    {
    }

	LRESULT OnKeyDown( UINT msg, WPARAM wParam,	LPARAM lParam, BOOL& isHandled )
	{
		if(wParam == VK_ESCAPE)
		{
			DestroyWindow();
		}
		return 0;
	}
	
	LRESULT OnDestroy( UINT msg, WPARAM wParam, LPARAM lParam, BOOL& isHandled )
    {
        PostQuitMessage(0);
        return 0;
    }

    LRESULT OnSize( UINT msg, WPARAM wParam, LPARAM lParam, BOOL& isHandled )
    {
        const DWORD width = LOWORD(lParam);
        const DWORD height = HIWORD(lParam);
        if( m_scene )
        {
            HRESULT hRes = m_scene->OnSize(width, height);
        }
        return 0;
    }

    LRESULT OnActivateApp( UINT msg, WPARAM wParam, LPARAM lParam, BOOL& isHandled )
    {
        bool activated = wParam != 0;

        if( m_scene )
        {
            HRESULT hRes = m_scene->OnActivateApp(activated);
        }
        return 0;
    }

    LRESULT OnDisplayModeChange( UINT msg, WPARAM wParam, LPARAM lParam, BOOL& isHandled )
    {
        if( m_scene )
        {
            HRESULT hRes = m_scene->OnDisplayModeChange();
        }
        return 0;
    }
	 
	bool ProcessMessage(WPARAM& uiRetCode)
    {
        MSG msg;
        ZeroMemory( &msg, sizeof(MSG) );
        while( ::PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE) )
        {
            if(WM_QUIT == msg.message)
            {
                uiRetCode = msg.wParam;
                return false;
            }
            ::TranslateMessage(&msg);
            ::DispatchMessage(&msg);
        }
        return true;
    }
};

//-----------------------------------------------------------------------------
// WinMain
//-----------------------------------------------------------------------------
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int )
{	
    // ATLW[
    CAtlWinModule module;
    {
        // IR[h
        UINT uiRetCode = 0;

        // COM 
        CoInitialize( NULL );

        // CEBhE 쐬
        MainWindow window;
        window.Create( NULL, CWindow::rcDefault,
            L"Execute Buffer Test", WS_CAPTION | WS_SYSMENU | WS_SIZEBOX | WS_MINIMIZEBOX);

        // sɂIʂ߂ return  break 邱
        do
        {
            window.ResizeClient( 640, 480, FALSE );
            window.CenterWindow();
            window.ShowWindow( SW_SHOW );

            HRESULT ret = E_FAIL;

            Scene scene;
			ret = scene.Initialize( window.m_hWnd );
            window.m_scene = &scene;

            if( !window.ProcessMessage(uiRetCode) )
            {
                break;
            }

            //// CbZ[W[v
            while( window.ProcessMessage(uiRetCode) )
            {
				ret = scene.AnimateScene();
				if( FAILED(ret) )
				{
					continue;
				}
				ret = scene.DoFrame();
				if( FAILED(ret) )
				{
					continue;
				}
            }
            // bZ[W[vI
			window.m_scene = NULL;
        }
		while(false);
		window.m_scene = NULL;

        CoUninitialize();

        return uiRetCode;
    }
}
